/*  DSP LIBRARY
**  
**  fft -- fft computation
**  
**  port of FFT Haskell : Copyright (c) 1998 Matthew P. Donadio (m.p.donadio@ieee.org)
**  GNU General public License terms apply
**
**  15/05/2001, Yacine Saidji
*/

def Fft = {

  val pi:double = 3.141592653589793238462643383279502884197169399375

  def flipflop[a] (x:a) ((e:List[a]),(o:List[a])) : (List[a],List[a]) =
    (List.cons[a] x o, e)

  def partitionEvenOdd[a] (xs:List[a]) : (List[a],List[a]) =
    xs.foldr (flipflop[a]) (List.empty[a],List.empty[a])

  // fixOrder creates a list with the even index elements in the beginning,
  // and the odd index elements in the end
  def fixOrder[a] (xs:List[a]) : List[a] =
    (val (e,o) = (partitionEvenOdd[a]) xs ; (List.append[a]) (e) (o))
  
  // fftCalc does the actual calculations
  // The y0s list represents the FFT of the even elements from the previous step
  // the y1s is the FFT of the odd elements
  // wn is the twiddle factor step, and w is the twiddle factor for the current step
  def fftCalc (y0s:List[Complex]) (y1s:List[Complex]) (wn:Complex) (w:Complex) : List[Complex] =
    if (y0s.isEmpty || y1s.isEmpty) List.empty[Complex]
    else (
      val y0 = y0s.lhead;
      val y1 = y1s.lhead;
      val tmp = w.mul(y1);
      List.cons[Complex] (y0.add (tmp)) 
      ( List.cons[Complex] (y0.sub (tmp)) 
       (fftCalc (y0s.ltail) (y1s.ltail) (wn) (wn.mul(w)) ))
      )


  // Dha FFT
  def fftcore (isign:int) (a:List[Complex]) : List[Complex] =
    if (a.ltail.isEmpty) a
    else {
      val n:double = (a.size).toDouble
      val wn = Complex.cis( (isign.toDouble) * (-2.0) * pi / n)
      val w = Complex.Complex(1.0,0.0)
      val (y0s,y1s) = partitionEvenOdd[Complex] (a)
      val y0 = fftcore (isign) (y0s)
      val y1 = fftcore (isign) (y1s)
      fixOrder[Complex] (fftCalc (y0) (y1) (wn) (w))
    }

  //  FFT
  def fft (a:List[Complex]) : List[Complex] = fftcore (1) (a)

  // inverse FFT (not scaled)
  def ifft (a:List[Complex]) : List[Complex] = fftcore (-1) (a)

  def scale (a:List[Complex]): List[Complex] =  {
    val n:double = (a.size).toDouble
    val ninv:Complex = Complex.Complex( (1.0) / n, 0.0 )
    a.map[Complex] (x:Complex | x.mul (ninv))
  }

}









